import json
import sys, os

import warnings
warnings.filterwarnings("ignore")
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)
import requests
from intelli_utilities import fetch_config_model, DATETIME_FORMATS, LOCAL_TIME_ZONE
from utils import get_model_params
from lib.Thresholds import Thresholds
from metrics.calculate_metrics import calculate_metrics
import pandas as pd
from datetime import datetime 

def main(args):
    input_uuid = args['inputDataUuid']
    token = args['token']
    process_manager = args['processManager']
    custom_payload = args['customPayload']

    config_model_uuid = custom_payload['configModelUuid']
    dependent_node_visualization_name = custom_payload['dependentNodeName']    

    with open('spool/{}/input.json'.format(input_uuid), 'r') as file:
        data = json.load(file)

    dataFrame = pd.DataFrame(data['dataFrame'])
    dataFrame = dataFrame.set_index('dateTime', drop=False)
    dataFrame.index = pd.to_datetime(dataFrame.index)                

    dataframes = []
    for date in custom_payload['trainingDates']:
        dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
        dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)
        dataframes.append(dataFrame[dt_start:dt_end])
    training_df = pd.concat(dataframes)

    transform_df = pd.DataFrame()
    if len(custom_payload['measureDates']) > 0:
        dataframes = []
        for date in custom_payload['measureDates']:
            dt_start = datetime.fromisoformat(date['start']).astimezone(LOCAL_TIME_ZONE)
            dt_end = datetime.fromisoformat(date['end']).astimezone(LOCAL_TIME_ZONE)  
            dataframes.append(dataFrame[dt_start:dt_end])
        transform_df = pd.concat(dataframes)

    config_model = {}
    outputs = {}
    thresholds = Thresholds()

    config_model = fetch_config_model(config_model_uuid, token, process_manager)
    thresholds.parse_config_model(config_model)

    dependent_node_df_colname = config_model['dependentVariables'][0]
    
    if 'skipComputation' not in config_model or config_model['skipComputation'] == False:            
        model_params = get_model_params(config_model)
        threshold_instance = thresholds.create_thresold_instance(config_model['task']['algorithm']['method'], config_model['dependentVariables'], config_model['independentVariables'], config_model['task']['algorithm']['model'], model_params)
        
        if config_model['trainMethod'] == 'movingWindows':         
            outputs = thresholds.do_backtesting(training_df, threshold_instance)

        if config_model['trainMethod'] == 'isolatedWindows':                
            outputs, last_trained_model = thresholds.do_single_train(threshold_instance, training_df, transform_df)                

            # Save last trained model
            # TODO Saved model is needed in compute_new_datapoints.
            # We could get rid of this and retrain the model in compute_new_datapoints.
            last_trained_model.save_model(config_model['uuid'], token, process_manager) 
    else:                    
        # Reset skipComputation
        config_model['skipComputation'] = False

    config_model['task']['metric'] = round(calculate_metrics(outputs['thresholds'], config_model['dependentVariables'][0]), 3)
    requests.post("{}/service/rest/v1/filedb/app/BAI/configModel/{}.json".format(process_manager, config_model['uuid']), json=config_model , headers={'Authorization': token }, verify=False)                                 
    
    thresholds_df = outputs['thresholds']
    thresholds_df = thresholds.set_event_flags(thresholds_df)
    thresholds_df = thresholds.set_alarm_flags(thresholds_df)
    thresholds_df = thresholds.bake_thresholds(thresholds_df)
    
    events_df = thresholds_df[(thresholds_df['event_upper'] == True) | (thresholds_df['event_lower'] == True)]
    alarms_df = thresholds_df[(thresholds_df['alarm_upper'] == True) | (thresholds_df['alarm_lower'] == True)]

    response = {
        'line': [
            { 
                'id': 'expectedValue',
                'name': 'expectedValue',
                'x': thresholds_df['dateTime'].values.tolist(),
                'y': thresholds_df['Y_pred'].values.tolist(),
                'color': config_model['expectedValueColor'],
                'yaxis': dependent_node_visualization_name,
                'zIndex': 50
            }
        ],
        'rangeArea': [
            {
                'id': 'thresholds',                
                'name': 'thresholds',
                'x': thresholds_df['ts'].values.tolist() ,
                'y': thresholds_df['lower'].values.tolist(),
                'y2': thresholds_df['upper'].values.tolist(),
                'color': config_model['thresholdColor'],
                'yaxis': dependent_node_visualization_name,
                'zIndex': 10
            }
        ], 
        'annotations': {
            'points': [
                {
                    'x': events_df['dateTime'].values.tolist(),
                    'y': events_df[dependent_node_df_colname].values.tolist(),
                    'color': 'blue',
                    'shape': 'circle',
                    'radius': 4,
                    'size': 5,
                    'yaxis': dependent_node_visualization_name
                },
                {
                    'x': alarms_df['dateTime'].values.tolist(),
                    'y': alarms_df[dependent_node_df_colname].values.tolist(),
                    'color': 'red',
                    'shape': 'circle',
                    'radius': 4,
                    'size': 5,
                    'yaxis': dependent_node_visualization_name
                },              
            ]
        },
        'thresholdsDataFrame' : thresholds_df.to_dict('records')
    }
    
    if 'outlier' in thresholds_df.columns: # outliers don't exist for isolatedWindows
        outlier_df =  thresholds_df[thresholds_df['outlier'] == True]
        response['annotations']['points'].append({
            'x': outlier_df['dateTime'].values.tolist(),
            'y': outlier_df[dependent_node_df_colname].values.tolist(),
            'color': 'black',
            'shape': 'square',
            'size': 5,
            'radius': 4,
            'yaxis': dependent_node_visualization_name
        })
        
    if config_model['trainMethod'] == 'movingWindows':
        response['lastRetrainDate'] = outputs['lastRetrainDate']
        response['trainingDates'] = outputs['trainingDates']
        response['reTrainingDates'] = outputs['reTrainingDates']
    
    response['lastThresholdDate'] = thresholds_df[-1:].index[0].strftime(DATETIME_FORMATS['local'])
    
    return response
       
if __name__ == "__main__":
    response = main(runpy_args)  
           
    #     if args['action'] == 'markerClick':
    #         clicked_datetime = list(main_hist_data[args['pointIndex']].keys())[0]
    #         clicked_datetime_obj = datetime.fromisoformat(clicked_datetime)
    #         clicked_timestamp = int(clicked_datetime_obj.timestamp() * 1000)
    #         xaxisAnnotationSurface = []
    #         opacity = 0.3
    #         for idx, retrain_ts in enumerate(outputs['retraining_dates']):
    #             if retrain_ts[0] <= clicked_timestamp and retrain_ts[1] >= clicked_timestamp:        
    #                 selected_train_ts = outputs['training_dates'][idx]
    #                 xaxisAnnotationSurface.append(create_xaxis_annotation('trainSurface', selected_train_ts[0], selected_train_ts[1], None, '#78b8f0', opacity))                                       
    #                 xaxisAnnotationSurface.append(create_xaxis_annotation('retrainSurface', retrain_ts[0], retrain_ts[1], None, '#f0c278', opacity))         
    #                 options['annotations']['xaxis'] = options['annotations']['xaxis'] + xaxisAnnotationSurface                

