
import os
import sys
from datetime import datetime, timedelta
import argparse
import pandas as pd
import json
import traceback
import numpy as np

# Load parent dir for intelli_utilities module
# Append maxpermissible values folder
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

# Append scripts folder
parentdir = os.path.dirname(parentdir)
sys.path.append(parentdir)

from intelli_utilities import fetch_widget, fetch_config_model, DATETIME_FORMATS, fetch_endpoint, fetch_change_log_list, serialize_interpolation, fetch_alarm_list
from utils import get_model_params
import requests
from lib.Thresholds import Thresholds
from DataProcessor import DataProcessor
from maxpermissiblevalues.metrics.metric_utils import get_node_unique_identifier

import logging
logging.basicConfig(filename='log/cron.log', encoding='utf-8', level=logging.ERROR)

# Three reasons to build dataframe in python 
# - Golang does not provide holidays calculation api!
# - Data amount to iterate shoud be much less
# - Allows to avoid configModel dependency in Action.go-> obsolete. Action has its own config model

BI_SYSTEM_PATHS = {
    'high': {
        'expected': 'BISystem:LimitHigh:{}:Expected',
        'limit': 'BISystem:LimitHigh:{}:Limit',
        'alarm': 'BISystem:LimitHigh:{}:Alarm',
        'pre_alarm': 'BISystem:LimitHigh:{}:PreAlarm',
    },
    'low': {
        'expected': 'BISystem:LimitLow:{}:Expected',
        'limit': 'BISystem:LimitLow:{}:Limit',
        'alarm': 'BISystem:LimitLow:{}:Alarm',
        'pre_alarm': 'BISystem:LimitLow:{}:PreAlarm',
    }
}

def set_trend_histdata(path:str, datetime:str, value:any) -> list:
    #Hist data attribute will create TRD points
    return {
            'path': path, 
            'type': 'double',
            'create': True,
            'value': value,
            'histData': [{
                datetime: value
            }],
    }


def get_protocol_template(path:str, create:bool, stamp:str, value:bool, type:str):
    return [
        {'path': path, 'type': 'bool', 'value': value, 'create': create, 'stamp': stamp},
        {'path': '{}:PRT'.format(path), 'type': 'none', 'value': None, 'create': create},
        {'path': '{}:PRT:Activ'.format(path), 'type': 'bool', 'value': True, 'create': create},
        {'path': '{}:PRT:Format'.format(path), 'type': 'string', 'value': type, 'create': create},
        {'path': '{}:PRT:Prot'.format(path), 'type': 'string', 'value': 'BISystem', 'create': create},
        {'path': '{}:PRT:Type'.format(path), 'type': 'int', 'value': 0, 'create': create},
    ]


def get_alarm_template(path:str, create:bool, stamp:str, value:any)  -> list:
    return [
        {'path': path, 'create': create,  'value': value, 'type': 'bool', 'stamp': stamp},
        {'path': '{}:ALM:Alarm1'.format(path), 'create': create, 'type': None},
        {'path': '{}:ALM:Alarm1:Activ'.format(path), 'create': create, 'value': True, 'type': 'bool'},
        {'path': '{}:ALM:Alarm1:ExtAlmText'.format(path), 'create': create, 'value': '', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Format'.format(path), 'create': create, 'type': 'string', 'value': ''},
        # {'path': '{}:ALM:Alarm1:Format'.format(path), 'create': create, 'value': 'Alarm1@Anlage / #c / #N / #V^NAME / #VComment / #Z(kommt:geht:quit) #u', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Group'.format(path), 'create': create, 'value': 0, 'type': 'int'},
        {'path': '{}:ALM:Alarm1:Pri'.format(path), 'create': create, 'value': 5, 'type': 'int'},
        {'path': '{}:ALM:Alarm1:Prot'.format(path), 'create': create, 'value': 'ALARM', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Text'.format(path), 'create': create, 'value': '', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Type'.format(path), 'create': create, 'value': '>=', 'type': 'string'},
        {'path': '{}:ALM:Alarm1:Value'.format(path), 'create': create, 'value': 1.000, 'type': 'double'},
    ]

def nodes_to_dataframe(nodes:list, interval, dates:list, dp:DataProcessor) -> pd.DataFrame:
    data_df = pd.DataFrame()
    for node in nodes:
        col_name = get_node_unique_identifier(node['pmModel'])
        hist_data = dp.get_hist_data_from_dates(node['pmModel'], node['interpolation'], interval, dates)
        df = dp.convert_histdata_to_dataframe(hist_data, col_name)
        # todo set holidays

        if 'bins' in node:
            if 'base' in node['bins'] and 'step' in node['bins']:
                df = dp.binarize_data(df, col_name, node['bins'])
        
        data_df = data_df.join(df, how='outer', sort=True)

    data_df = dp.add_temporal_variables(data_df)
    return data_df

def persist_config_model(config_model:dict, args):
    # persist next training date
    uri = '{}/service/rest/v1/filedb/app/BAI/configModel/{}.json'.format(args.process_manager, config_model['uuid'])
    requests.post(uri,
        json=config_model, 
        headers={'Authorization': args.token},
        verify=False            
    )

def get_training_dates(config_model:dict, now:datetime):
    training_dates = []
    match config_model['trainMethod']['type']:
        case 'movingWindows':
            start_date = now - timedelta(days=config_model['trainMethod']['periodsDay']['train'])
            training_dates = [{
                'start': start_date.strftime(DATETIME_FORMATS['local']),
                'end': now.strftime(DATETIME_FORMATS['local'])
            }]                            
        case 'isolatedWindows':
            training_dates = config_model['trainMethod']['trainingDates']
        case _:
            raise Exception('trainMethod does not exist')
    return training_dates

def is_counter(interpolation:dict):
    counter = False
    if 'template' in interpolation:
        if interpolation['template'] in ['energyCounterUp', 'energyCounterUpDown']:
            counter = True

    if 'method' in interpolation:
        if interpolation['method']['diff'] in ['diff', 'diffCtrUpCleaned']:
            counter = True

    return counter

def get_transform_df(config_model:dict, now, token:str, process_manager:str):
    """
    Fetches the current datapoint value
    For counters, the use the last value and the latest one the hour value to perform differentiation
    """
    start_date = None
    match config_model['interval']:
        case 3600:
            start_date = now - timedelta(hours=1)
        case -1:
            start_date = now - timedelta(days=1)
        case _:
            raise Exception('interval does not exist')

    start_date_time = start_date.strftime(DATETIME_FORMATS['local'])
    end_date_time = now.strftime(DATETIME_FORMATS['local'])

    data_df = pd.DataFrame()
    current_value = None

    for node in config_model['nodes']:
        col_name = get_node_unique_identifier(node['pmModel'])
        endpoint = fetch_endpoint(node['pmModel'], False, token, process_manager)

        payload = {}
        if is_counter(node['interpolation']):
            payload = {
            "get": [
                { # fetches on the hour value
                    "path": node['pmModel']['path'],
                    "showExtInfos": ["name", "unit"],
                    "histData": {
                        'start': start_date_time,
                        'end': end_date_time,
                        'interval': config_model['interval'],
                        'interpolateMethod': 'prevNextLinearFill'
                    }
                },
                {# fetches last value
                    "path": node['pmModel']['path'],
                    "showExtInfos": ["name", "unit"],
                    "histData": {
                        'start': start_date_time,
                        'end': end_date_time,
                        'interpolateMethod': 'prev'
                    }
                },            
            ]}
            response = requests.post(endpoint, data=json.dumps(payload), headers={
                'Content-Type': 'application/json',
                'Authorization': token
            })
            response.raise_for_status()
            data = json.loads(response.text)

            latest_value = list(data['get'][1]['histData'][-1].values())[0]
            on_the_hour_value = list(data['get'][0]['histData'][0].values())[0]
            current_value = latest_value - on_the_hour_value

            for i, hist in enumerate(data['get'][0]['histData']):
                logging.error('On the hour {} {}: {}'.format(i, list(hist.values())[0], list(hist.keys())[0]))
                print('On the hour {} {}: {}'.format(i, list(hist.values())[0], list(hist.keys())[0]))

            for i, hist in enumerate(data['get'][1]['histData']):
                logging.error('latest value {} {}: {}'.format(i, list(hist.values())[0], list(hist.keys())[0]))
                print('latest value {} {}: {}'.format(i, list(hist.values())[0], list(hist.keys())[0]))

            logging.error('final values')
            logging.error('on the hour: {}'.format(on_the_hour_value))
            logging.error('latest: {}'.format(latest_value))
            logging.error('current: {}'.format(current_value))

        else:
            payload = {
                "get": [
                    {
                        "path": node['pmModel']['path'],
                        "showExtInfos": ["name", "unit"],
                        "histData": {
                            'start': start_date_time,
                            'end': end_date_time,
                        }
                    },            
                ]}
            if 'template' in node['interpolation'] and node['interpolation']['template'] not in ['', 'custom']:                
                payload['get'][0]['histData']['template'] = node['interpolation']['template'] 
            else:
                payload['get'][0]['histData']['interpolateMethod'] = serialize_interpolation(node['interpolation']['method'])


            response = requests.post(endpoint, data=json.dumps(payload), headers={
                'Content-Type': 'application/json',
                'Authorization': token
            })
            response.raise_for_status()
            data = json.loads(response.text)
            current_value = list(data['get'][0]['histData'][-1].values())[0]
            logging.error('transform undiffed value {}'.format(np.round(current_value, 2)))


        df = pd.DataFrame({
            col_name: np.round(current_value, 2)
        }, index=[now])

        if 'bins' in node:
            if 'base' in node['bins'] and 'step' in node['bins']:
                df = dp.binarize_data(df, col_name, node['bins'])
        data_df = data_df.join(df, how='outer', sort=True)

    data_df = dp.add_temporal_variables(data_df)
    # todo set holidays
    return data_df

if __name__ == '__main__':
    # Set cli args
    parser = argparse.ArgumentParser()
    parser.add_argument('--widget-uuid')
    parser.add_argument('--action-uuid')
    parser.add_argument('--token', help='Security token for services')
    parser.add_argument('--process-manager', help='Base url of the process manager')
    parser.add_argument('--now')
    args = parser.parse_args()


    logging.error('----- START DEBUGGER -------')
    logging.error('executed at {}'.format(args.now))
    logging.error('widget uuid {}'.format(args.widget_uuid))
    logging.error('action uuid {}'.format(args.action_uuid))

    try:
        now = datetime.strptime(args.now, '%Y-%m-%dT%H:%M:%S%z')
        dp = DataProcessor(args.token, args.process_manager)

        whois = 'BI'
        user = 'bi_cron'

        try:
            widget = fetch_widget(args.widget_uuid, args.token, args.process_manager)
            action = None
            for item in widget['actions']:
                if item['uuid'] == args.action_uuid:
                    action = item
            config_model = fetch_config_model(widget['configModelUuid'], args.token, args.process_manager)
        except requests.exceptions.RequestException as e:
            raise SystemExit(e)
        
        # pick dependent var pmModel
        dependent_pm_model = None
        for node in config_model['nodes']:
            unique_name = get_node_unique_identifier(node['pmModel'])
            if unique_name == config_model['dependentVariables'][0]:
                dependent_pm_model = node['pmModel']
                break    

        logging.error('actionType: {}'.format(action['type']))
        logging.error('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['pre_alarm'].format(config_model['pathId'])))
            
        thresholds = Thresholds()
        thresholds.parse_action_config_model(config_model)
        # Todo set model parameters! model_params = get_model_params(config_model)    
        threshold_instance = thresholds.create_thresold_instance(config_model['algorithm']['method'], config_model['dependentVariables'], config_model['independentVariables'], config_model['algorithm']['model'])    
        try:
            threshold_instance.load_model(action['modelUuid'], args.token, args.process_manager)                            
        except FileNotFoundError:
            training_dates = get_training_dates(config_model, now)

            if config_model['trainMethod']['type'] == 'movingWindows':
                next_training_date = now + timedelta(days=config_model['trainMethod']['periodsDay']['retrain'])
                config_model['trainMethod']['nextTraining'] = next_training_date.strftime(DATETIME_FORMATS['local'])            
                persist_config_model(config_model, args)

            train_df = nodes_to_dataframe(config_model['nodes'], config_model['interval'], training_dates, dp)
            # Train and persist model
            threshold_instance.fit(train_df)
            threshold_instance.save_model(action['modelUuid'], args.token, args.process_manager)

        if config_model['trainMethod']['type'] == 'movingWindows':
            next_training_date = datetime.strptime(config_model['trainMethod']['nextTraining'], DATETIME_FORMATS['local']) 
            if now > next_training_date:
                training_dates = get_training_dates(config_model, now)

                next_training_date = now + timedelta(days=config_model['trainMethod']['periodsDay']['retrain'])
                config_model['trainMethod']['nextTraining'] = next_training_date.strftime(DATETIME_FORMATS['local'])            
                persist_config_model(config_model, args)

                train_df = nodes_to_dataframe(config_model['nodes'], config_model['interval'], training_dates, dp)
                # Train and persist model
                threshold_instance.fit(train_df)
                threshold_instance.save_model(action['modelUuid'], args.token, args.process_manager)            


        data_df = get_transform_df(config_model, now, args.token, args.process_manager)    
        result = threshold_instance.transform(data_df)

        # Read events from change logs
        upper_event_count = config_model['eventCount']['upper']
        lower_event_count = config_model['eventCount']['lower']

        delta_param_upper = {}
        delta_param_lower = {}
        alarm_delta = {}
        match config_model['interval']:
            case 3600:
                delta_param_upper = {'hours': upper_event_count}
                delta_param_lower = {'hours': lower_event_count}
                alarm_delta = {'hours': 1}
            case -1:
                delta_param_upper = {'days': upper_event_count}
                delta_param_lower = {'days': lower_event_count}
                alarm_delta = {'days': 1}
            case _:
                raise Exception('interval not found')

        start_date_upper = (now - timedelta(**delta_param_upper)).strftime(DATETIME_FORMATS['local'])
        start_date_lower = (now - timedelta(**delta_param_lower)).strftime(DATETIME_FORMATS['local'])

        endpoint = fetch_endpoint(dependent_pm_model, False, args.token, args.process_manager)
        change_logs_upper = fetch_change_log_list(endpoint, start_date_upper, now.strftime(DATETIME_FORMATS['local']), args.token, '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['pre_alarm'].format(config_model['pathId'])))
        change_logs_lower = fetch_change_log_list(endpoint, start_date_lower, now.strftime(DATETIME_FORMATS['local']), args.token, '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['pre_alarm'].format(config_model['pathId'])))

        alarm_start = (now - timedelta(**alarm_delta)).strftime(DATETIME_FORMATS['local'])
        alarm_upper_list = fetch_alarm_list(endpoint, alarm_start, now.strftime(DATETIME_FORMATS['local']), args.token, '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['alarm'].format(config_model['pathId'])))
        alarm_lower_list = fetch_alarm_list(endpoint, alarm_start, now.strftime(DATETIME_FORMATS['local']), args.token, '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['alarm'].format(config_model['pathId'])))

        #Filter out activated changelogs
        change_logs_upper = [log for log in change_logs_upper if log['newValue'] == '1']
        change_logs_lower = [log for log in change_logs_lower if log['newValue'] == '1']

        logging.error("len(change_logs_upper): {}".format(len(change_logs_upper)))
        logging.error("len(change_logs_lower): {}".format(len(change_logs_lower)))

        logging.error("len(alarm_upper): {}".format(len(alarm_upper_list)))
        logging.error("len(alarm_lower): {}".format(len(alarm_lower_list)))  

        logging.error("dist_upper: {}".format(result['dist_upper'].values[0]))
        logging.error("dist_lower: {}".format(result['dist_lower'].values[0]))

        logging.error("Y_pred: {}".format(result['Y_pred'].values[0]))
        logging.error("th_upper: {}".format(result['upper'].values[0]))
        logging.error("th_lower: {}".format(result['lower'].values[0]))

        event_upper = False
        alarm_upper = False
        if result['dist_upper'].values[0] > 0:
            if len(alarm_upper_list) > 0 or len(change_logs_upper) >= upper_event_count:
                alarm_upper = True            
            else:
                event_upper = True
            

        event_lower = False
        alarm_lower = False
        if result['dist_lower'].values[0] > 0 or len(change_logs_lower) >= lower_event_count:
            if len(alarm_lower_list) > 0:
                alarm_lower = True            
            else:
                event_lower = True

        logging.error("event_upper: {}".format(event_upper))
        logging.error("event_lower: {}".format(event_lower))
        logging.error("alarm_lower: {}".format(alarm_lower))
        logging.error("alarm_upper: {}".format(alarm_upper))


        now_str = now.strftime(DATETIME_FORMATS['local'])

        # Hanlde preAlarm and alarms for plotting purpose with help of the type
        enable_change_log = []
        disable_change_log = []    
        if event_upper == True or alarm_upper == True:
            upper_type = 'preAlarm' if event_upper == True else 'alarm'
            enable_change_log = enable_change_log + get_protocol_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['pre_alarm'].format(config_model['pathId'])), True, now_str, True, upper_type) 
            disable_change_log = disable_change_log + get_protocol_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['pre_alarm'].format(config_model['pathId'])), True, now_str, False, upper_type)

        if event_lower == True or alarm_lower == True:
            lower_type = 'preAlarm' if event_lower == True else 'alarm'
            enable_change_log = enable_change_log + get_protocol_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['pre_alarm'].format(config_model['pathId'])), True, now_str, True, lower_type) 
            disable_change_log = disable_change_log + get_protocol_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['pre_alarm'].format(config_model['pathId'])), True, now_str, False, lower_type) 


        # Handle only alarms. Has no effect on alarm plotting
        set_alarm = []
        if alarm_lower or alarm_upper:
            if alarm_lower:
                set_alarm = set_alarm + get_alarm_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['alarm'].format(config_model['pathId'])), True, now_str, alarm_lower)
            if alarm_upper:
                set_alarm = set_alarm + get_alarm_template('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['alarm'].format(config_model['pathId'])), True, now_str, alarm_upper)
        
        if alarm_upper == False and event_upper == False:
           set_alarm.append({'path': '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['alarm'].format(config_model['pathId'])), 'type': 'bool', 'value': False})
        if alarm_lower == False and event_lower == False:
           set_alarm.append({'path': '{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['alarm'].format(config_model['pathId'])), 'type': 'bool', 'value': False})           
            
        response = []
        match action['type']:
            case 'trend':
                response.append({
                    'type': 'trend',
                    'reqDms': [
                        {
                            'payload': {
                                'whois': whois,
                                'user': user,
                                'set': [
                                    set_trend_histdata('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['expected'].format(config_model['pathId'])), now_str, result['Y_pred'].values[0]),
                                    set_trend_histdata('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['expected'].format(config_model['pathId'])), now_str, result['Y_pred'].values[0]),
                                    set_trend_histdata('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['high']['limit'].format(config_model['pathId'])), now_str, result['upper'].values[0]),
                                    set_trend_histdata('{}:{}'.format(dependent_pm_model['path'], BI_SYSTEM_PATHS['low']['limit'].format(config_model['pathId'])), now_str, result['lower'].values[0]),
                                ]
                            },                
                            'pmModel': dependent_pm_model
                        },                    
                    ]
                })
            case 'alarm':
                response.append({
                'type': 'alarm',
                'reqDms': [
                    {
                        'payload': {                
                            'whois': whois,
                            'user': user,
                            'set': set_alarm                           
                        },
                        'pmModel': dependent_pm_model
                    },
                    {
                        'payload': {                
                            'whois': whois,
                            'user': user,
                            'set': enable_change_log   
                        },
                        'pmModel': dependent_pm_model
                    },
                    {
                        'payload': {                
                            'whois': whois,
                            'user': user,
                            'set': disable_change_log  
                        },
                        'pmModel': dependent_pm_model
                    }                                            
                ]
            })

        print(json.dumps(response))
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]        
        error = "{} {} {} {} {}".format(e, exc_type, fname, exc_tb.tb_lineno, traceback.format_exc())
        logging.error(error)